from numpy.linalg import pinv
from numpy import (
    all,
    asarray,
    atleast_2d,
    clip,
    copyto,
    empty,
    full,
    inf,
    isfinite,
    log,
    newaxis,
    sqrt,
)

from .._util import cached_property, hinv, hsolve, log2pi, nice_inv, rsolve, safe_log
from ._b import B


class FastScanner:
    """
    Approximated fast inference over several covariates.

    Specifically, it maximizes the marginal likelihood ::

        p(𝐲)ⱼ = 𝓝(𝐲 | 𝚇𝜷ⱼ + 𝙼ⱼ𝜶ⱼ, 𝑠ⱼ(𝙺 + 𝑣𝙸)),

    over 𝜷ⱼ, 𝜶ⱼ, and sⱼ. Matrix Mⱼ is the candidate set defined by the user. Variance 𝑣
    is not optimised for performance reasons. The method assumes the user has provided a
    reasonable value for it.

    Parameters
    ----------
    y
        Real-valued outcome.
    X
        Matrix of covariates.
    QS
        Economic eigendecomposition ``((Q0, Q1), S0)`` of ``K``.
    v
        Variance due to iid effect.

    Notes
    -----
    The implementation requires further explanation as it is somehow obscure. Let
    𝚀𝚂𝚀ᵀ = 𝙺, where 𝚀𝚂𝚀ᵀ is the eigendecomposition of 𝙺. Let 𝙳 = (𝚂 + 𝑣𝙸) and
    𝙳₀ = (𝚂₀ + 𝑣𝙸₀), where 𝚂₀ is the part of 𝚂 with positive values. Therefore, solving

        (𝙺 + 𝑣𝙸)𝐱 = 𝐲

     for 𝐱 is equivalent to solving

        𝚀₀𝙳₀𝚀₀ᵀ𝐱 + 𝑣𝚀₁𝚀₁ᵀ𝐱 = 𝚀₀𝙳₀𝚀₀ᵀ𝐱 + 𝑣(𝙸 - 𝚀₀𝚀₀ᵀ)𝐱 = 𝐲.

    for 𝐱. Let

        𝙱 = 𝚀₀𝙳₀⁻¹𝚀₀ᵀ                    if 𝑣=0, and
        𝙱 = 𝚀₀𝙳₀⁻¹𝚀₀ᵀ + 𝑣⁻¹(𝙸 - 𝚀₀𝚀₀ᵀ)   if 𝑣>0.

    We therefore have

        𝐱 = 𝙱𝐲

    as the solution of (𝙺 + 𝑣𝙸)𝐱 = 𝐲.

    Let 𝐛ⱼ = [𝜷ⱼᵀ 𝜶ⱼᵀ]ᵀ and 𝙴ⱼ = [𝚇 𝙼ⱼ]. The optimal parameters according to the marginal
    likelihood are given by ::

        (𝙴ⱼᵀ𝙱𝙴ⱼ)𝐛ⱼ = 𝙴ⱼᵀ𝙱𝐲

    and ::

        𝑠 = 𝑛⁻¹𝐲ᵀ𝙱(𝐲 - 𝙴ⱼ𝐛ⱼ).
    """

    def __init__(self, y, X, QS, v):
        from numpy_sugar import epsilon

        y = asarray(y, float)
        X = atleast_2d(asarray(X, float).T).T

        if not all(isfinite(y)):
            raise ValueError("Not all values are finite in the outcome array.")

        if not all(isfinite(X)):
            raise ValueError("Not all values are finite in the `X` matrix.")

        if v < 0:
            raise ValueError("Variance has to be non-negative.")

        if not isfinite(v):
            raise ValueError("Variance has to be a finite value..")

        if v <= epsilon.small:
            v = 0.0

        Q0 = QS[0][0]
        S0 = QS[1]
        D0 = S0 + v
        self._rankdef = y.shape[0] - S0.shape[0]

        self._B = B(Q0, S0, 1.0, v)

        By = self._B.dot(y)
        self._yTBy = y.T @ By
        self._yTBX = By.T @ X

        # Used for performing association scan on single variants
        self._ETBE = ETBE(self._B, X)
        self._yTBE = yTBE(By.T @ X)

        self._Q0 = Q0
        self._D0 = D0
        self._X = X
        self._y = y
        self._v = v

    def null_lml(self) -> float:
        return self._null_lml

    @cached_property
    def _null_lml(self) -> float:
        """
        Log of the marginal likelihood for the null hypothesis.

        It is implemented as ::

            2·log(p(𝐲)) = -𝑛·log(2·𝜋·𝑠) - log|D| - 𝑛.

        Returns
        -------
        lml
            Log of the marginal likelihood.
        """
        n = self._nsamples
        scale = self.null_scale
        return (self._static_lml - n * log(scale)) / 2

    @cached_property
    def null_beta(self):
        """
        Optimal 𝜷 according to the marginal likelihood.

        It is compute by solving the equation ::

            (XᵀBX)𝜷 = XᵀB𝐲.

        Returns
        -------
        beta
            Optimal 𝜷.
        """
        return rsolve(self._ETBE.XTBX, self._yTBX)

    @cached_property
    def null_beta_covariance(self):
        """
        Covariance of the optimal 𝜷 according to the marginal likelihood.

        Returns
        -------
        beta_covariance
            (Xᵀ(s(K + vI))⁻¹X)⁻¹.
        """
        return self.null_scale * nice_inv(self._X.T @ self._B.dot(self._X))

    @cached_property
    def null_beta_se(self):
        """
        Standard errors of the optimal 𝜷.

        Returns
        -------
        beta_se
            Square root of the diagonal of the beta covariance.
        """
        return sqrt(self.null_beta_covariance.diagonal())

    @cached_property
    def null_scale(self) -> float:
        """
        Optimal s according to the marginal likelihood.

        The optimal s is given by ::

            s = n⁻¹𝐲ᵀB(𝐲 - X𝜷),

        where 𝜷 is optimal.

        Returns
        -------
        scale
            Optimal scale.
        """
        n = self._nsamples
        beta = self.null_beta
        sqrdot2 = self._yTBy - self._yTBX @ beta
        return sqrdot2 / n

    def fast_scan(self, M, verbose: bool = True):
        """
        LMLs, fixed-effect sizes, and scales for single-marker scan.

        Parameters
        ----------
        M : array_like
            Matrix of fixed-effects across columns.
        verbose
            ``True`` for progress information; ``False`` otherwise.
            Defaults to ``True``.

        Returns
        -------
        lmls
            Log of the marginal likelihoods.
        effsizes0
            Covariate fixed-effect sizes.
        effsizes1
            Candidate set fixed-effect sizes.
        scales
            Scales.
        """
        from tqdm import tqdm

        if M.ndim != 2:
            raise ValueError("`M` array must be bidimensional.")
        p = M.shape[1]

        lmls = empty(p)
        effsizes0 = empty((p, self._X.shape[1]))
        effsizes0_se = empty((p, self._X.shape[1]))
        effsizes1 = empty(p)
        effsizes1_se = empty(p)
        scales = empty(p)

        chunks = get_chunks(M)

        start = 0
        for i in tqdm(chunks, desc="Scanning", disable=not verbose):
            stop = start + i

            r = self._fast_scan_chunk(M[:, start:stop])

            lmls[start:stop] = r["lml"]
            effsizes0[start:stop, :] = r["effsizes0"]
            effsizes0_se[start:stop, :] = r["effsizes0_se"]
            effsizes1[start:stop] = r["effsizes1"]
            effsizes1_se[start:stop] = r["effsizes1_se"]
            scales[start:stop] = r["scale"]

            start = stop

        return {
            "lml": lmls,
            "effsizes0": effsizes0,
            "effsizes0_se": effsizes0_se,
            "effsizes1": effsizes1,
            "effsizes1_se": effsizes1_se,
            "scale": scales,
        }

    def scan(self, M):
        """
        LML, fixed-effect sizes, and scale of the candidate set.

        Parameters
        ----------
        M : array_like
            Fixed-effects set.

        Returns
        -------
        lml : float
            Log of the marginal likelihood.
        effsizes0 : ndarray
            Covariates fixed-effect sizes.
        effsizes0_se : ndarray
            Covariates fixed-effect size standard errors.
        effsizes1 : ndarray
            Candidate set fixed-effect sizes.
        effsizes1_se : ndarray
            Candidate fixed-effect size standard errors.
        scale : ndarray
            Optimal scale.
        """
        from numpy_sugar import is_all_finite

        M = asarray(M, float)

        if M.shape[1] == 0:
            return {
                "lml": self._null_lml,
                "effsizes0": self.null_beta,
                "effsizes0_se": self.null_beta_se,
                "effsizes1": empty((0)),
                "effsizes1_se": empty((0)),
                "scale": self.null_scale,
            }

        if not is_all_finite(M):
            raise ValueError("M parameter has non-finite elements.")

        BM = self._B.dot(M)
        yTBM = self._y.T @ BM
        XTBM = self._X.T @ BM
        MTBM = M.T @ BM

        return self._multicovariate_set(yTBM, XTBM, MTBM)

    @property
    def _nsamples(self):
        return self._y.shape[0]

    @property
    def _ncovariates(self):
        return self._X.shape[1]

    @cached_property
    def _static_lml(self):
        """
        Static part of the marginal likelihood.

        It is defined by ::

            -𝑛·log(2·𝜋) - 𝑛 - log|D|.
        """
        n = self._nsamples
        static_lml = -n * log2pi - n
        static_lml -= safe_log(self._D0).sum()
        static_lml -= self._rankdef * safe_log(self._v)
        return static_lml

    def _fast_scan_chunk(self, M):
        from numpy_sugar import dotd

        M = asarray(M, float)

        if not M.ndim == 2:
            raise ValueError("`M` array must be bidimensional.")

        if not all(isfinite(M)):
            raise ValueError("One or more variants have non-finite value.")

        BM = self._B.dot(M)
        yTBM = self._y.T @ BM
        XTBM = self._X.T @ BM
        dMTBM = dotd(M.T, BM)

        lmls = full(M.shape[1], self._static_lml)
        eff0 = empty((M.shape[1], self._X.shape[1]))
        eff0_se = empty((M.shape[1], self._X.shape[1]))
        eff1 = empty((M.shape[1]))
        eff1_se = empty((M.shape[1]))
        scales = empty(M.shape[1])

        effs = {"eff0": eff0, "eff0_se": eff0_se, "eff1": eff1, "eff1_se": eff1_se}

        if self._ncovariates == 1:
            self._1covariate_loop(lmls, effs, scales, yTBM, XTBM, dMTBM)
        else:
            self._multicovariate_loop(lmls, effs, scales, yTBM, XTBM, dMTBM)

        return {
            "lml": lmls,
            "effsizes0": eff0,
            "effsizes0_se": eff0_se,
            "effsizes1": eff1,
            "effsizes1_se": eff1_se,
            "scale": scales,
        }

    def _multicovariate_loop(self, lmls, effs, scales, yTBM, XTBM, diagMTBM):
        ETBE = self._ETBE
        yTBE = self._yTBE

        for i in range(XTBM.shape[1]):

            yTBE.set_yTBM(yTBM[i])
            ETBE.set_XTBM(XTBM[:, [i]])
            ETBE.set_MTBM(diagMTBM[i])

            left = ETBE.value
            right = yTBE.value
            x = rsolve(left, right)
            beta = x[:-1][:, newaxis]
            alpha = x[-1:]
            bstar = _bstar_unpack(beta, alpha, self._yTBy, yTBE, ETBE, bstar_1effect)

            se = sqrt(nice_inv(left).diagonal())

            scales[i] = bstar / self._nsamples
            lmls[i] -= self._nsamples * safe_log(scales[i])
            effs["eff0"][i, :] = beta.T
            effs["eff0_se"][i, :] = se[:-1]
            effs["eff1"][i] = alpha[0]
            effs["eff1_se"][i] = se[-1]

        lmls /= 2

    def _multicovariate_set(self, yTBM, XTBM, MTBM):

        yBE = yTBE(self._yTBX, yTBM.shape[0])
        yBE.set_yTBM(yTBM)

        set_size = yTBM.shape[0]
        EBE = ETBE(self._B, self._X, set_size)

        EBE.set_XTBM(XTBM)
        EBE.set_MTBM(MTBM)

        left = EBE.value
        right = yBE.value
        x = rsolve(left, right)

        beta = x[:-set_size]
        alpha = x[-set_size:]
        bstar = _bstar_unpack(beta, alpha, self._yTBy, yBE, EBE, bstar_set)

        lml = self._static_lml

        scale = bstar / self._nsamples
        lml -= self._nsamples * safe_log(scale)
        lml /= 2

        effsizes_se = sqrt(scale * nice_inv(left).diagonal())
        beta_se = effsizes_se[:-set_size]
        alpha_se = effsizes_se[-set_size:]

        return {
            "lml": lml,
            "effsizes0": beta,
            "effsizes0_se": beta_se,
            "effsizes1": alpha,
            "effsizes1_se": alpha_se,
            "scale": scale,
        }

    def _1covariate_loop(self, lmls, effs, scales, yTBM, XTBM, diagMTBM):
        ETBE = self._ETBE
        yTBX = self._yTBX
        XTBX = ETBE.XTBX
        yTBy = self._yTBy

        A00 = ETBE.XTBX[0, 0]
        A01 = XTBM[0, :]
        A11 = diagMTBM

        b0 = yTBX[0]
        b1 = yTBM

        x = hsolve(A00, A01, A11, b0, b1)
        beta = x[0][newaxis, :]
        alpha = x[1]
        bstar = bstar_1effect(beta, alpha, yTBy, yTBX, yTBM, XTBX, XTBM, diagMTBM)

        scales[:] = bstar / self._nsamples
        lmls -= self._nsamples * safe_log(scales)
        lmls /= 2
        effs["eff0"][:] = beta.T
        effs["eff1"][:] = alpha

        A00i, _, A11i = hinv(A00, A01, A11)
        effs["eff0_se"][:, 0] = sqrt(scales * A00i)
        effs["eff1_se"][:] = sqrt(scales * A11i)


class yTBE:
    """
    Represent 𝐲ᵀ𝙱𝙴 where 𝙴 = [𝚇 𝙼].
    """

    def __init__(self, yTBX, set_size=1):
        n = yTBX.shape[0] + set_size
        self._data = empty((n,))
        self._data[:-set_size] = yTBX
        self._m = set_size

    @property
    def value(self):
        return self._data

    @property
    def yTBX(self):
        return self._data[: -self._m]

    @property
    def yTBM(self):
        return self._data[-self._m :]

    def set_yTBM(self, yTBM):
        copyto(self.yTBM, yTBM)


class ETBE:
    """
    Represent 𝙴ᵀ𝙱𝙴 where 𝙴 = [𝚇 𝙼].
    """

    def __init__(self, B, X, set_size=1):
        n = X.shape[1] + set_size
        BX = B.dot(X)
        self._data = empty((n, n))
        self._data[:-set_size, :-set_size] = X.T @ BX
        self._m = set_size

    @property
    def value(self):
        return self._data

    @property
    def XTBX(self):
        return self._data[: -self._m, : -self._m]

    @property
    def XTBM(self):
        return self._data[: -self._m, -self._m :]

    @property
    def MTBX(self):
        return self._data[-self._m :, : -self._m]

    @property
    def MTBM(self):
        return self._data[-self._m :, -self._m :]

    def set_XTBM(self, XTBM):
        copyto(self.XTBM, XTBM)
        copyto(self.MTBX, XTBM.T)

    def set_MTBM(self, MTBM):
        copyto(self.MTBM, MTBM)


def bstar_1effect(beta, alpha, yTBy, yTBX, yTBM, XTBX, XTBM, MTBM):
    """
    Same as :func:`bstar_set` but for single-effect.
    """
    from numpy_sugar import epsilon
    from numpy_sugar.linalg import dotd

    r = full(MTBM.shape[0], yTBy)
    r -= 2 * yTBX @ beta
    r -= 2 * yTBM * alpha
    r += dotd(beta.T, XTBX @ beta)
    r += dotd(beta.T, XTBM * alpha)
    r += (alpha * XTBM * beta).sum(axis=0)
    r += alpha * MTBM.ravel() * alpha
    return clip(r, epsilon.tiny, inf)


def bstar_set(beta, alpha, yTBy, yTBX, yTBM, XTBX, XTBM, MTBM):
    """
    Compute 𝐲ᵀ𝙱𝐲 - 2𝐲ᵀ𝙱𝙴ⱼ𝐛ⱼ + 𝐛ⱼᵀ𝙴ⱼᵀ𝙱𝙴ⱼ𝐛ⱼ.

    For 𝐛ⱼ = [𝜷ⱼᵀ 𝜶ⱼᵀ]ᵀ.
    """
    r = yTBy
    r -= 2 * yTBX @ beta
    r -= 2 * yTBM @ alpha
    r += beta.T @ XTBX @ beta
    r += 2 * beta.T @ XTBM @ alpha
    r += alpha.T @ MTBM @ alpha
    return r


def _bstar_unpack(beta, alpha, yTBy, yTBE, ETBE, bstar):
    from numpy_sugar import epsilon

    yTBX = yTBE.yTBX
    yTBM = yTBE.yTBM
    XTBX = ETBE.XTBX
    XTBM = ETBE.XTBM
    MTBM = ETBE.MTBM
    bstar = bstar(beta, alpha, yTBy, yTBX, yTBM, XTBX, XTBM, MTBM)
    return clip(bstar, epsilon.tiny, inf).flat[0]


def get_chunks(M):
    chunks = None
    if hasattr(M, "chunks") and M.chunks is not None:
        if len(M.chunks) == 2:
            return M.chunks[1]

    p = M.shape[1]
    siz = round(p / min(50, p))
    n = int(p / siz)
    chunks = [siz] * n
    if n * siz < p:
        chunks += [p - n * siz]

    return chunks



from numpy import array, concatenate, eye, kron
from numpy.linalg import slogdet
from numpy.random import default_rng
from numpy.testing import assert_allclose
from scipy.optimize import check_grad

from glimix_core._util import vec
from glimix_core.cov import Kron2SumCov


def test_kron2sumcov():
    G = array([[-1.5, 1.0], [-1.5, 1.0], [-1.5, 1.0]])
    Lr = array([[3], [2]], float)
    Ln = array([[1, 0], [2, 1]], float)

    cov = Kron2SumCov(G, 2, 1)
    cov.C0.L = Lr
    cov.C1.L = Ln

    I = eye(G.shape[0])
    assert_allclose(
        cov.value(), kron(Lr @ Lr.T, G @ G.T) + kron(Ln @ Ln.T, I), atol=1e-4
    )
    assert_allclose(cov._check_grad(), 0, atol=1e-5)
    assert_allclose(cov.solve(cov.value()), eye(2 * G.shape[0]), atol=1e-7)
    assert_allclose(cov.logdet(), slogdet(cov.value())[1], atol=1e-7)

    def func(x):
        cov.C0.Lu = x[:2]
        cov.C1.Lu = x[2:]
        return cov.logdet()

    def grad(x):
        cov.C0.Lu = x[:2]
        cov.C1.Lu = x[2:]
        D = cov.logdet_gradient()
        return concatenate((D["C0.Lu"], D["C1.Lu"]))

    random = default_rng(0)
    assert_allclose(check_grad(func, grad, random.normal(size=5)), 0, atol=1e-5)

    V = random.normal(size=(3, 2))

    g = cov.C0.gradient()["Lu"]
    g0 = cov.gradient_dot(vec(V))["C0.Lu"]
    for i in range(2):
        assert_allclose(g0[..., i], kron(g[..., i], G @ G.T) @ vec(V))

    g = cov.C1.gradient()["Lu"]
    g0 = cov.gradient_dot(vec(V))["C1.Lu"]
    for i in range(3):
        assert_allclose(g0[..., i], kron(g[..., i], eye(3)) @ vec(V))

    V = random.normal(size=(3, 2, 4))

    g = cov.C0.gradient()["Lu"]
    g0 = cov.gradient_dot(vec(V))["C0.Lu"]
    for i in range(2):
        for j in range(4):
            assert_allclose(g0[j, ..., i], kron(g[..., i], G @ G.T) @ vec(V[..., j]))

    g = cov.C1.gradient()["Lu"]
    g0 = cov.gradient_dot(vec(V))["C1.Lu"]
    for i in range(3):
        for j in range(4):
            assert_allclose(g0[j, ..., i], kron(g[..., i], eye(3)) @ vec(V[..., j]))

    M = random.normal(size=(2 * G.shape[0], 2 * 4))

    R = cov.LdKL_dot(M)
    dK = cov.gradient()
    L = kron(cov.Lh, cov.Lx)

    for i in range(cov.C0.shape[0]):
        for j in range(M.shape[1]):
            expected = L @ dK["C0.Lu"][..., i] @ L.T @ M[:, [j]]
            assert_allclose(R["C0.Lu"][:, [j], i], expected, atol=1e-7)

    for i in range(cov.C1.shape[0]):
        for j in range(M.shape[1]):
            expected = L @ dK["C1.Lu"][..., i] @ L.T @ M[:, [j]]
            assert_allclose(R["C1.Lu"][:, [j], i], expected, atol=1e-7)


def test_kron2sumcov_g_full_col_rank():
    G = array([[-1.5, 1.0, 0.2, 0.5], [1.0, -0.25, -1.5, 1.0], [-0.1, -0.20, -2.5, 0]])
    Lr = array([[3], [2]], float)
    Ln = array([[1, 0], [2, 1]], float)

    cov = Kron2SumCov(G, 2, 1)
    cov.C0.L = Lr
    cov.C1.L = Ln

    I = eye(G.shape[0])
    assert_allclose(
        cov.value(), kron(Lr @ Lr.T, G @ G.T) + kron(Ln @ Ln.T, I), atol=1e-4
    )
    assert_allclose(cov._check_grad(), 0, atol=1e-5)
    assert_allclose(cov.solve(cov.value()), eye(2 * G.shape[0]), atol=1e-7)
    assert_allclose(cov.logdet(), slogdet(cov.value())[1], atol=1e-7)

    cov = Kron2SumCov(G, 2, 2)
    Lr = array([[3, 0.0], [-2, 0.4]], float)
    Ln = array([[1, 0], [2, 1]], float)
    cov.C0.L = Lr
    cov.C1.L = Ln

    I = eye(G.shape[0])
    assert_allclose(
        cov.value(), kron(Lr @ Lr.T, G @ G.T) + kron(Ln @ Ln.T, I), atol=1e-4
    )
    assert_allclose(cov._check_grad(), 0, atol=1e-5)
    assert_allclose(cov.solve(cov.value()), eye(2 * G.shape[0]), atol=1e-7)
    assert_allclose(cov.logdet(), slogdet(cov.value())[1], atol=1e-7)



from numpy.random import default_rng


class GGPSampler(object):
    r"""Sample from a Generalised Gaussian Process.

    Outcome modelled via

    .. math::

        \mathbf y \sim \int \prod_i \mathrm{ExpFam}(y_i ~|~ g_i(z_i))
        \mathcal N(\mathbf z ~|~ \mathbf m; \mathrm K) \mathrm d\mathbf z.

    Parameters
    ----------
    link : str
        Likelihood name.
    mean : function
        Mean function.
    cov : function
        Covariance function.

    Example
    -------

    .. doctest::

        >>> from numpy.random import default_rng
        >>>
        >>> from glimix_core.example import offset_mean
        >>> from glimix_core.example import linear_eye_cov
        >>> from glimix_core.random import GGPSampler
        >>> from glimix_core.lik import DeltaProdLik
        >>>
        >>> random = default_rng(1)
        >>>
        >>> mean = offset_mean()
        >>> cov = linear_eye_cov()
        >>>
        >>> lik = DeltaProdLik()
        >>>
        >>> y = GGPSampler(lik, mean, cov).sample(random)
        >>> print(y[:5])  # doctest: +FLOAT_CMP
        [-2.42181498  0.50720447 -1.01053967  0.736624    1.64019063]
    """

    def __init__(self, lik, mean, cov):
        self._lik = lik
        self._mean = mean
        self._cov = cov

    def sample(self, random_state=None):
        r"""Sample from the specified distribution.

        Parameters
        ----------
        random_state : random_state
            Set the initial random state.

        Returns
        -------
        numpy.ndarray
            Sample.
        """
        from numpy_sugar import epsilon
        from numpy_sugar.linalg import sum2diag
        from numpy_sugar.random import multivariate_normal

        if random_state is None:
            random_state = default_rng()

        m = self._mean.value()
        K = self._cov.value().copy()

        sum2diag(K, +epsilon.small, out=K)

        return self._lik.sample(multivariate_normal(m, K, random_state), random_state)



import pytest
from numpy import array, asarray, block, inf, nan, ones
from numpy.linalg import inv, pinv, svd
from numpy.testing import assert_allclose
from numpy_sugar.linalg import ddot

from glimix_core._util import check_covariates, check_economic_qs, check_outcome, hinv
from glimix_core._util.solve import heigvals, hsvd


def test_util_check_economic_qs():
    A = ones((3, 2))
    B = ones((3, 1))
    C = ones(2)

    with pytest.raises(ValueError):
        check_economic_qs(A)

    with pytest.raises(ValueError):
        check_economic_qs((A, C))

    A[0, 0] = inf
    QS = ((A, B), C)

    with pytest.raises(ValueError):
        check_economic_qs(QS)

    A[0, 0] = 1
    C[0] = nan

    with pytest.raises(ValueError):
        check_economic_qs(QS)


def test_util_check_covariates():
    A = ones(2)
    B = ones((1, 2))

    with pytest.raises(ValueError):
        check_covariates(A)

    B[0, 0] = inf
    with pytest.raises(ValueError):
        check_covariates(B)


def test_util_check_outcome():
    y = ones(5)

    y[0] = nan
    with pytest.raises(ValueError):
        check_outcome((y,), "poisson")

    y[0] = 0.5
    want = array([0.5, 1.0, 1.0, 1.0, 1.0])
    assert_allclose(check_outcome(y, "poisson"), want)

    x = ones(4)

    with pytest.raises(ValueError):
        check_outcome((y, x), "bernoulli")

    x = ones(5)

    with pytest.raises(ValueError):
        check_outcome(y, "normal")


def test_util_check_poisson_outcome():
    y = ones(5)
    y[0] = 25000 + 1
    want = array(
        [2.50000000e04, 1.00000000e00, 1.00000000e00, 1.00000000e00, 1.00000000e00]
    )
    with pytest.warns(UserWarning):
        assert_allclose(check_outcome(y, "poisson"), want)


def test_hsvd():
    A = asarray([[1.2, -0.2], [-0.2, 1.1]])
    SVD0 = hsvd(A[0, 0], A[0, 1], A[1, 1])
    SVD0 = [block(i) for i in SVD0]
    SVD1 = svd(A)
    assert_allclose(SVD0[0] @ ddot(SVD0[1], SVD0[2]), SVD1[0] @ ddot(SVD1[1], SVD1[2]))

    A = asarray([[0, 0], [0, 0]])
    SVD0 = hsvd(A[0, 0], A[0, 1], A[1, 1])
    SVD0 = [block(i) for i in SVD0]
    SVD1 = svd(A)
    assert_allclose(SVD0[0] @ ddot(SVD0[1], SVD0[2]), SVD1[0] @ ddot(SVD1[1], SVD1[2]))

    A = asarray([[1, 0], [0, 0]])
    SVD0 = hsvd(A[0, 0], A[0, 1], A[1, 1])
    SVD0 = [block(i) for i in SVD0]
    SVD1 = svd(A)
    assert_allclose(SVD0[0] @ ddot(SVD0[1], SVD0[2]), SVD1[0] @ ddot(SVD1[1], SVD1[2]))


def test_hinv():
    for A00 in [1.3, -1.3]:
        A01 = asarray([0.2, 0.1, 0.15, -0.3, 1.299, -0.1])
        A11 = asarray([0.9, 1.1, 0.75, 1.3, 1.3, 1000000.0])
        ai, bi, di = hinv(A00, A01, A11)
        for i in range(len(bi)):
            Ai = inv([[A00, A01[i]], [A01[i], A11[i]]])
            assert_allclose([[ai[i], bi[i]], [bi[i], di[i]]], Ai, atol=1e-5, rtol=1e-5)

    A = asarray([[0.3119092627599252, 0.0], [0.0, 0.0]])
    ai, bi, di = hinv(A[0, 0], A[0, 1:], A[1, 1:])
    assert_allclose([[ai[0], bi[0]], [bi[0], di[0]]], pinv(A), atol=1e-5, rtol=1e-5)

    A = asarray([[0.311909262759, 0.311909262759], [0.311909262759, 0.311909262759]])
    ai, bi, di = hinv(A[0, 0], A[0, 1:], A[1, 1:])
    assert_allclose([[ai[0], bi[0]], [bi[0], di[0]]], pinv(A), atol=1e-5, rtol=1e-5)

    A = [asarray([[0.3119092627599252, 0.0], [0.0, 0.0]])]
    A += [asarray([[0.311909262759, 0.311909262759], [0.311909262759, 0.311909262759]])]
    A00 = [0.3119092627599252, 0.311909262759]
    A01 = [0.0, 0.311909262759]
    A11 = [0.0, 0.311909262759]

    for i in range(2):
        ai, bi, di = hinv(A00[i], A01[i], A11[i])
        ai = ai[0]
        bi = bi[0]
        di = di[0]
        assert_allclose([[ai, bi], [bi, di]], pinv(A[i]), atol=1e-5, rtol=1e-5)


def test_heigvals():
    a = 1.0
    b = [1.0, 0.0, 100000000000.0, -0.03]
    d = [1.0, 0.0, 1.0, -0.4]
    heigvals(a, b, d)

    a = 0.0
    b = [1.0, 0.0, 100000000000.0, -0.03]
    d = [1.0, 0.0, 1.0, -0.4]
    heigvals(a, b, d)



import pytest
from numpy.random import default_rng
from numpy.testing import assert_allclose, assert_array_less, assert_equal

from glimix_core.cov import EyeCov, LinearCov, SumCov
from glimix_core.lik import BinomialProdLik, PoissonProdLik
from glimix_core.link import LogitLink, LogLink
from glimix_core.mean import OffsetMean
from glimix_core.random import (
    GGPSampler,
    GPSampler,
    bernoulli_sample,
    binomial_sample,
    poisson_sample,
)


def test_binomial_sampler():
    random = default_rng(4503)
    link = LogitLink()
    binom = BinomialProdLik(12, link)
    assert_equal(binom.sample(0, random), 4.0)


def test_poisson_sampler():
    random = default_rng(4503)
    link = LogLink()
    poisson = PoissonProdLik(link)
    assert_equal(poisson.sample(0, random), 0.0)
    assert_equal(poisson.sample(0, random), 1.0)
    assert_equal(poisson.sample(0, random), 2.0)
    assert_equal(poisson.sample(0, random), 2.0)
    assert_equal(poisson.sample(-10, random), 0.0)
    assert_equal(poisson.sample(+5, random), 160.0)


@pytest.mark.skip(reason="non-deterministic")
def test_GGPSampler_poisson():
    random = default_rng(4503)
    X = random.normal(size=(10, 15))
    link = LogLink()
    lik = PoissonProdLik(link)

    mean = OffsetMean(10)
    mean.offset = 1.2
    cov = LinearCov(X)
    sampler = GGPSampler(lik, mean, cov)
    assert_equal(
        sampler.sample(random), [1.0, 0.0, 0.0, 10.0, 2.0, 618.0, 17.0, 3.0, 20.0, 9.0]
    )

    mean = OffsetMean(10)
    mean.offset = 0.0

    cov1 = LinearCov(X)

    cov2 = EyeCov(10)

    cov1.scale = 1e-4
    cov2.scale = 1e-4

    cov = SumCov([cov1, cov2])

    sampler = GGPSampler(lik, mean, cov)

    assert_equal(
        sampler.sample(random), [2.0, 2.0, 0.0, 3.0, 0.0, 7.0, 1.0, 1.0, 1.0, 1.0]
    )

    cov2.scale = 20.0
    sampler = GGPSampler(lik, mean, cov)
    assert_equal(
        sampler.sample(random), [12.0, 0.0, 29.0, 0.0, 0.0, 0.0, 38.0, 5.0, 0.0, 0.0]
    )

    sampler.sample()


@pytest.mark.skip(reason="non-deterministic")
def test_GGPSampler_binomial():
    random = default_rng(4503)
    X = random.normal(size=(10, 15))
    link = LogitLink()
    lik = BinomialProdLik(5, link)

    mean = OffsetMean(10)
    mean.offset = 1.2
    cov = LinearCov(X)
    sampler = GGPSampler(lik, mean, cov)
    assert_equal(
        sampler.sample(random), [2.0, 1.0, 3.0, 5.0, 4.0, 5.0, 5.0, 4.0, 5.0, 5.0]
    )

    mean.offset = 0.0
    assert_equal(
        sampler.sample(random), [0.0, 3.0, 1.0, 0.0, 5.0, 4.0, 0.0, 4.0, 1.0, 1.0]
    )

    mean = OffsetMean(10)
    mean.offset = 0.0

    cov1 = LinearCov(X)

    cov2 = EyeCov(10)

    cov1.scale = 1e-4
    cov2.scale = 1e-4

    cov = SumCov([cov1, cov2])

    lik = BinomialProdLik(100, link)
    sampler = GGPSampler(lik, mean, cov)
    assert_equal(
        sampler.sample(random),
        [40.0, 39.0, 50.0, 47.0, 56.0, 61.0, 54.0, 54.0, 50.0, 52.0],
    )

    cov2.scale = 100.0
    sampler = GGPSampler(lik, mean, cov)
    assert_equal(
        sampler.sample(random),
        [98.0, 0.0, 100.0, 44.0, 100.0, 90.0, 100.0, 99.0, 22.0, 16.0],
    )


def test_canonical_bernoulli_sampler():
    random = default_rng(9)
    G = random.normal(size=(10, 5))

    y = bernoulli_sample(0.1, G, random_state=random)
    assert_array_less(y, [2] * 10)


def test_canonical_binomial_sampler():
    random = default_rng(9)
    G = random.normal(size=(10, 5))

    y = binomial_sample(5, 0.1, G, random_state=random)
    assert_array_less(y, [5 + 1] * 10)

    ntrials = [2, 3, 1, 1, 4, 5, 1, 2, 1, 1]
    y = binomial_sample(ntrials, -0.1, G, random_state=random)
    assert_array_less(y, [i + 1 for i in ntrials])

    X = random.normal(size=len(ntrials))
    y = binomial_sample(ntrials, -0.1, G, causal_variants=X, random_state=random)
    assert_array_less(y, [i + 1 for i in ntrials])

    X = random.normal(size=(len(ntrials), 2))
    y = binomial_sample(ntrials, -0.1, G, causal_variants=X, random_state=random)
    assert_array_less(y, [i + 1 for i in ntrials])


def test_canonical_poisson_sampler():
    random = default_rng(10)
    G = random.normal(size=(10, 5))

    y = poisson_sample(0.1, G, random_state=random)
    assert_array_less(y, [20] * len(y))


@pytest.mark.skip(reason="non-deterministic")
def test_GPSampler():
    random = default_rng(4503)
    X = random.normal(size=(10, 15))

    mean = OffsetMean(10)
    mean.offset = 1.2
    cov = LinearCov(X)
    sampler = GPSampler(mean, cov)

    x = [
        -0.9328779739,
        -1.997157809,
        0.5276438923,
        2.3730973762,
        1.2032887445,
        6.4469744126,
        2.7004521738,
        0.5853558975,
        2.8682819171,
        2.2248157174,
    ]
    assert_allclose(sampler.sample(random), x)
    sampler.sample()



from numpy import asarray, dot, ones, zeros
from optimix import Function, Vector

from .._util import format_function


class LRFreeFormCov(Function):
    """
    General semi-definite positive matrix of low rank, K = LLᵀ.

    The covariance matrix K is given by LLᵀ, where L is a n×m matrix and n≥m. Therefore,
    K will have rank(K) ≤ m.

    Example
    -------

    .. doctest::

        >>> from glimix_core.cov import LRFreeFormCov
        >>> cov = LRFreeFormCov(3, 2)
        >>> print(cov.L)
        [[1. 1.]
         [1. 1.]
         [1. 1.]]
        >>> cov.L = [[1, 2], [0, 3], [1, 3]]
        >>> print(cov.L)
        [[1. 2.]
         [0. 3.]
         [1. 3.]]
        >>> cov.name = "F"
        >>> print(cov)
        LRFreeFormCov(n=3, m=2): F
          L: [[1. 2.]
              [0. 3.]
              [1. 3.]]
    """

    def __init__(self, n, m):
        """
        Constructor.

        Parameters
        ----------
        n : int
            Covariance dimension.
        m : int
            Upper limit of the covariance matrix rank.
        """
        self._L = ones((n, m))
        self._Lu = Vector(self._L.ravel())
        Function.__init__(self, "LRFreeFormCov", Lu=self._Lu)

    @property
    def nparams(self):
        """
        Number of parameters.
        """
        return self._L.size

    def listen(self, func):
        """
        Listen to parameters change.

        Parameters
        ----------
        func : callable
            Function to be called when a parameter changes.
        """
        self._Lu.listen(func)

    def fix(self):
        """
        Disable parameter optimisation.
        """
        self._Lu.fix()

    def unfix(self):
        """
        Enable parameter optimisation.
        """
        self._Lu.unfix()

    @property
    def Lu(self):
        """
        Lower-triangular, flat part of L.
        """
        return self._Lu.value

    @Lu.setter
    def Lu(self, v):
        self._Lu.value = v

    @property
    def L(self):
        """
        Matrix L from K = LLᵀ.

        Returns
        -------
        L : (n, m) ndarray
            Parametric matrix.
        """
        return self._L

    @L.setter
    def L(self, value):
        self._Lu.value = asarray(value, float).ravel()

    @property
    def shape(self):
        """
        Array shape.
        """
        n = self._L.shape[0]
        return (n, n)

    def value(self):
        """
        Covariance matrix.

        Returns
        -------
        K : (n, n) ndarray
            K = LLᵀ.
        """
        return dot(self.L, self.L.T)

    def gradient(self):
        """
        Derivative of the covariance matrix over the lower triangular, flat part of L.

        It is equal to

            ∂K/∂Lᵢⱼ = ALᵀ + LAᵀ,

        where Aᵢⱼ is an n×m matrix of zeros except at [Aᵢⱼ]ᵢⱼ=1.

        Returns
        -------
        Lu : ndarray
            Derivative of K over the lower-triangular, flat part of L.
        """
        L = self.L
        n = self.L.shape[0]
        grad = {"Lu": zeros((n, n, n * self._L.shape[1]))}
        for ii in range(self._L.shape[0] * self._L.shape[1]):
            row = ii // self._L.shape[1]
            col = ii % self._L.shape[1]
            grad["Lu"][row, :, ii] = L[:, col]
            grad["Lu"][:, row, ii] += L[:, col]

        return grad

    def __str__(self):
        return format_function(
            self, {"n": self._L.shape[0], "m": self._L.shape[1]}, [("L", self._L)]
        )



import pytest
from numpy import eye, inf, nan, zeros
from numpy.random import default_rng
from numpy.testing import assert_allclose

from glimix_core.cov import EyeCov, GivenCov, LinearCov, SumCov
from glimix_core.gp import GP
from glimix_core.mean import LinearMean, OffsetMean


def test_gp_value_1():
    random = default_rng(94584)
    N = 50
    X = random.normal(size=(N, 100))
    offset = 0.5

    mean = OffsetMean(N)
    mean.offset = offset
    mean.fix_offset()

    cov = LinearCov(X)
    cov.scale = 1.0

    y = random.normal(size=N)

    gp = GP(y, mean, cov)
    assert_allclose(gp.value(), -153.5086866719)


def test_gp_value_2():
    random = default_rng(94584)
    N = 50
    X1 = random.normal(size=(N, 3))
    X2 = random.normal(size=(N, 100))

    mean = LinearMean(X1)

    cov = LinearCov(X2)
    cov.scale = 1.0

    y = random.normal(size=N)

    gp = GP(y, mean, cov)
    assert_allclose(gp.value(), -154.1531720155)

    mean.effsizes = [3.4, 1.11, -6.1]
    assert_allclose(gp.value(), -174.088721839)


def test_gp_gradient():
    random = default_rng(94584)
    N = 50
    X = random.normal(size=(N, 100))
    offset = 0.5

    mean = OffsetMean(N)
    mean.offset = offset
    mean.fix_offset()

    cov = LinearCov(X)
    cov.scale = 1.0

    y = random.normal(size=N)

    gp = GP(y, mean, cov)

    assert_allclose(gp._check_grad(), 0, atol=1e-5)


def test_gp_gradient2():
    random = default_rng(94584)
    N = 10
    X = random.normal(size=(N, 2))

    K0 = random.normal(size=(N, N + 1))
    K0 = K0 @ K0.T + eye(N) * 1e-5

    K1 = random.normal(size=(N, N + 1))
    K1 = K1 @ K1.T + eye(N) * 1e-5

    mean = LinearMean(X)
    mean.effsizes = [-1.2, +0.9]

    cov0 = GivenCov(K0)
    cov1 = GivenCov(K1)
    cov2 = EyeCov(N)
    cov = SumCov([cov0, cov1, cov2])
    cov0.scale = 0.1
    cov1.scale = 2.3
    cov2.scale = 0.3

    y = random.normal(size=N)

    gp = GP(y, mean, cov)

    assert_allclose(gp._check_grad(), 0, atol=1e-5)


def test_gp_maximize():
    random = default_rng(94584)
    N = 50
    X = random.normal(size=(N, 100))
    offset = 0.5

    mean = OffsetMean(N)
    mean.offset = offset
    mean.fix_offset()

    cov = LinearCov(X)
    cov.scale = 1.0

    y = random.normal(size=N)

    gp = GP(y, mean, cov)

    assert_allclose(gp.value(), -153.5086866719)
    gp.fit(verbose=False)
    assert_allclose(gp.value(), -84.9518606572)


def test_gp_nonfinite_phenotype():
    random = default_rng(94584)
    N = 50
    X = random.normal(size=(N, 100))
    offset = 0.5

    mean = OffsetMean(N)
    mean.offset = offset
    mean.fix_offset()

    cov = LinearCov(X)
    cov.scale = 1.0

    y = zeros(N)

    y[0] = nan
    with pytest.raises(ValueError):
        GP(y, mean, cov)

    y[0] = -inf
    with pytest.raises(ValueError):
        GP(y, mean, cov)

    y[0] = +inf
    with pytest.raises(ValueError):
        GP(y, mean, cov)



from numpy import corrcoef, dot, ones, sqrt
from numpy.random import default_rng
from numpy.testing import assert_allclose
from numpy_sugar.linalg import economic_qs_linear

from glimix_core.cov import EyeCov, LinearCov, SumCov
from glimix_core.lik import DeltaProdLik
from glimix_core.lmm import LMM, LMMPredict
from glimix_core.mean import OffsetMean
from glimix_core.random import GGPSampler


def test_lmm_predict():
    random = default_rng(9458)
    n = 30

    X = random.normal(size=(n, n + 1))
    X -= X.mean(0)
    X /= X.std(0)
    X /= sqrt(X.shape[1])

    offset = 1.0

    mean = OffsetMean(n)
    mean.offset = offset

    cov_left = LinearCov(X)
    cov_left.scale = 1.5

    cov_right = EyeCov(n)
    cov_right.scale = 1.5

    cov = SumCov([cov_left, cov_right])

    lik = DeltaProdLik()

    y = GGPSampler(lik, mean, cov).sample(random)

    QS = economic_qs_linear(X)

    lmm = LMM(y, ones((n, 1)), QS)

    lmm.fit(verbose=False)

    plmm = LMMPredict(y, lmm.beta, lmm.v0, lmm.v1, lmm.mean(), lmm.covariance())

    K = dot(X, X.T)
    pm = plmm.predictive_mean(ones((n, 1)), K, K.diagonal())
    assert_allclose(corrcoef(y, pm)[0, 1], 0.966467656)
